;;;; NOTE: You should probably be looking at pnm.scm, not this file.  
;;;; I've included it just for fun.  This was my first implementation and
;;;; is C-style.  It immediately imports the entire file unlike the newer
;;;; version that uses streams more elegantly.

;;; Takes an input-port f and returns a list of strings representing 
;;; the tokens in f.  An ASCII PNM-style file consists of numbers and
;;; (some) special character strings separated by whitespace.  "#" 
;;; delimits end-of-line comments.
;;;
;;; Note: the implementation here is a bit C-style and with a little
;;; more thought could probably be made a little more Scheme-like.
(define (tokenize-pnm f)  
  ;; The current token is stored as a reversed list of characters 
  ;; and the list of completed tokens is stored as a reversed
  ;; list of strings.  If we had chosen to build these up in
  ;; non-reversed order, our algorithm would be Theta(n^2) instead
  ;; of Theta(n).
  (let ((reversed-tokens '())
        (current-token '())
        (*whitespace* '(#\Return #\Linefeed #\Tab #\Space)))
        
    ; Takes the current-token and prepends it to reversed-tokens,
    ; if it has any characters in it (we don't support zero-length
    ; tokens).  The current-token is first transformed into a
    ; non-reversed string.
    (define (merge-token)
      (if (not (null? current-token))
          (begin
            ;(display  (list->string (reverse current-token))) (display " ")
            (set! reversed-tokens (cons (list->string (reverse current-token))
                                        reversed-tokens))
            (set! current-token '()))))
    
    ; Appends (prepends actually) a new character c to the 
    ; current-token.
    (define (append-to-token c)
      (set! current-token (cons c current-token)))
    
    ; Loop through all input characters
    (let loop ()
      ; Read one in and process it...
      (let ((c (input-port/read-char f)))
        (cond 
          ; Regular whitespace: merge in the current token, if needed
          ((memq c *whitespace*) 
           (merge-token)
           (loop))
          ; Comment: slurp up all characters until (and including)
          ; the end of line.
          ((eq? c #\#) 
           (merge-token)
           (let comment-slurper ((c c))
             (if (not (eq? c #\Linefeed)) 
                 (comment-slurper (input-port/read-char f))))
           (loop))
          ; End of file: return the de-reversed tokens
          ((eof-object? c) 
           (reverse reversed-tokens))
          ; Else: accumulate characters
          (else 
           (append-to-token c) 
           (loop)))))))

;;; Reads a PPM file into a stream format, where the stream-car
;;; part of the stream is an image row.  Each image row is a
;;; stream where its stream-car part is an RGB triple, where
;;; each value is an integer between 0 and 1.
;;;
;;; http://netpbm.sourceforge.net/doc/ppm.html
(define (ppm-file->stream fname)
  (let* ((tokens       (tokenize-pnm (open-input-file fname)))
         (magic-number (first tokens))
         (width        (string->number (second tokens)))
         (height       (string->number (third tokens)))
         (max-value    (string->number (fourth tokens))))
    ;; Make sure we have an actual PPM file
    (assert-equal magic-number "P3")
    (assert= (length tokens) (+ 4 (* 3 width height)))
    (let row-loop ((data (cddddr tokens))
                   (num-rows height))
      (if (<= num-rows 0) '()
          (cons-stream
           (let col-loop ((data data) (num-cols width))
             (if (<= num-cols 0) '()
                 (cons-stream (map (lambda (x) (/ x max-value))
                                   (map string->number (sublist data 0 3)))
                              (col-loop (list-tail data 3) (dec num-cols)))))
           (row-loop (list-tail data (* 3 width)) (dec num-rows)))))))

;;; Reads a PGM file into a stream format, where the stream-car
;;; part of the stream is an image row.  Each image row is a
;;; stream where its stream-car part is a pixel gray value, from
;;; 0 to 1.
;;;
;;; http://netpbm.sourceforge.net/doc/pgm.html
(define (pgm-file->image fname)
  (let* ((tokens       (tokenize-pnm (open-input-file fname)))
         (magic-number (first tokens))
         (width        (string->number (second tokens)))
         (height       (string->number (third tokens)))
         (max-value    (string->number (fourth tokens))))
    ;; Make sure we have an actual PGM file
    (assert-equal magic-number "P2")
    (assert= (length tokens) (+ 4 (* width height)))
    (let row-loop ((data (cddddr tokens))
                   (num-rows height))
      (if (<= num-rows 0) '()
          (cons-stream
           (let col-loop ((data data) (num-cols width))
             (if (<= num-cols 0) '()
                 (cons-stream (/ (string->number (car data)) max-value)
                              (col-loop (cdr data) (dec num-cols)))))
           (row-loop (list-tail data width) (dec num-rows)))))))
           
(define (image->pgm-file img fname)
  (define (write-rows img f)
    (if (not (null? img)) 
        (begin
          (write-cols (stream-car img) f)
          (write-rows (stream-cdr img) f))))
  
  (define (write-cols row f)
    (if (null? row)
        (write-string "\n" f)
        (begin
          (write-string (number->string 
                         (inexact->exact (floor (* (stream-car row) 255.)))) f) 
          (write-string " " f)
          (write-cols (stream-cdr row) f))))               
  
  (let ((f      (open-output-file fname))
        (height (stream-length img))
        (width  (stream-length (stream-car img))))
    (write-string "P2\n" f)
    (write-string 
     (apply string-append
            `(,(number->string width) " " ,(number->string height) "\n255\n")) 
     f)
    (write-rows img f)
    (close-port f)))